/*
        sw_view.c

	player eye positioning

        Copyright (C) 1996-1997  Id Software, Inc.

        This program is free software; you can redistribute it and/or
        modify it under the terms of the GNU General Public License
        as published by the Free Software Foundation; either version 2
        of the License, or (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

        See the GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to:

                Free Software Foundation, Inc.
                59 Temple Place - Suite 330
                Boston, MA  02111-1307, USA

        $Id$
*/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "quakedef.h"
#include "r_local.h"

extern  byte    gammatable[256];

void V_CalcPowerupCshift (void);
qboolean V_CheckGamma (void);

/*
=============
V_UpdatePalette
=============
*/
void V_UpdatePalette (void)
{
        int             i, j;
        qboolean        new;
        byte    *basepal, *newpal;
        byte    pal[768];
        int             r,g,b;
        qboolean force;

        V_CalcPowerupCshift ();

        new = false;

        for (i=0 ; i<NUM_CSHIFTS ; i++)
        {
                if (cl.cshifts[i].percent != cl.prev_cshifts[i].percent)
                {
                        new = true;
                        cl.prev_cshifts[i].percent = cl.cshifts[i].percent;
                }
                for (j=0 ; j<3 ; j++)
                        if (cl.cshifts[i].destcolor[j] != cl.prev_cshifts[i].destcolor[j])
                        {
                                new = true;
                                cl.prev_cshifts[i].destcolor[j] = cl.cshifts[i].destcolor[j];
                        }
        }

// drop the damage value
        cl.cshifts[CSHIFT_DAMAGE].percent -= host_frametime*150;
        if (cl.cshifts[CSHIFT_DAMAGE].percent <= 0)
                cl.cshifts[CSHIFT_DAMAGE].percent = 0;
// drop the bonus value
        cl.cshifts[CSHIFT_BONUS].percent -= host_frametime*100;
        if (cl.cshifts[CSHIFT_BONUS].percent <= 0)
                cl.cshifts[CSHIFT_BONUS].percent = 0;

        force = V_CheckGamma ();
        if (!new && !force)
                return;

        basepal = host_basepal;
        newpal = pal;

        for (i=0 ; i<256 ; i++)
        {
                r = basepal[0];
                g = basepal[1];
                b = basepal[2];
                basepal += 3;

                for (j=0 ; j<NUM_CSHIFTS ; j++)
                {
                        r += (cl.cshifts[j].percent*(cl.cshifts[j].destcolor[0]-r))>>8;
                        g += (cl.cshifts[j].percent*(cl.cshifts[j].destcolor[1]-g))>>8;
                        b += (cl.cshifts[j].percent*(cl.cshifts[j].destcolor[2]-b))>>8;
                }

                newpal[0] = gammatable[r];
                newpal[1] = gammatable[g];
                newpal[2] = gammatable[b];
                newpal += 3;
        }

        VID_ShiftPalette (pal);
}
